/*
    Ishtar - A network transparent read/write interface with
    probing and enumeration support.
    service part : network transparent access to generic services
    Copyright (C) 2003-2005 Stephane Magnenat <nct@ysagoon.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
	Services defintion

	All things to have services working
*/

#ifndef __ISHTAR_SERVICES_H
#define __ISHTAR_SERVICES_H

#include <ishtar/network.h>

#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include <map>
#include <set>
#include <list>
#include <assert.h>

/*!	\file services.h
	\brief Interface of Ishtar services classes.
*/

/*!	\mainpage Ishtar services library reference documentation

	\section intro Introduction
	Ishtar is a network toolbox.
	Ishtar is composed of three differents modules.
	This modules is libishtarservices. Built upon libishtarnet, it implements a protocol for remote procedure call (RPC) and provides basic bricks for building such procedures.
	
	\section installation Installation
	The installation procedure is explained step by step in the \ref InstallGuide.
	
	\section usage Usage
	A quick start to most important classes is given in the \ref Cookbook.
	
	\section feedback Feedback
	If you have any comments or suggestions, do not hesitate to send them to
	Stephane Magnenat <nct@ysagoon.com>
	
	\section license License
	This program is free software released under the GNU General Public License version 2.
*/

//! Ishtar is the namespace of the Ishtar class library
namespace Ishtar
{
	//! Ishtar services protocol version, only programs with the same protocol version can communication together
	const UInt32 PROTOCOL_VERSION = 2;
	
	// Vector value handling
	// =====================
	
	//! A vector of scalar values
	struct ValueVector
	{
		//! Type of scalar values
		enum Type
		{
			//! 32 bits signed integer, corresponds to Int32
			INT=0,
			//! 32 bits unsigned integer, corresponds to UInt32
			UINT,
			//! 16 bits signed integer, corresponds to Int16
			SHORT,
			//! 16 bits unsigned integer, corresponds to UInt16
			USHORT,
			//! 8 bits signed integer, corresponds to Int8
			CHAR,
			//! 8 bits unsigned integer, corresponds to UInt8
			UCHAR,
			//! 32 bits floating point value, corresponds to float
			FLOAT,
			//! 64 bits double precision floating point value, corresponds to double
			DOUBLE,
			//! true/false value, serialized on a UCHAR
			BOOL
		};
	
		//! Destructor
		virtual ~ValueVector() { }
	
		//! Serialize this vector to a stream, write the type that call serializeSpecified
		virtual void serialize(OutputStream *stream) const = 0;
		//! Serialize the values of this vector to a stream, write the length then each value
		virtual void serializeSpecified(OutputStream *stream) const = 0;
		//! Deserialize the values of this vector form a stream, the type has already been read because the vector is already created
		virtual void deserializeSpecified(InputStream *stream) = 0;
	
		//! Get value index to a string
		virtual std::string toString(unsigned index) const = 0;
		//! Set value index from a string s
		virtual void fromString(unsigned index, const std::string &s) = 0;
	
		//! Return the type of the vector
		virtual Type getType(void) const = 0;
		//! Return the length of the vector
		virtual Size size(void) const = 0;
	
		//! Create a new ValueVector of Type type
		static ValueVector *newOfType(Type type, Size size = 0);
		//! Deserialize a vector from a stream and return it. First read the type, then create a new vector with newOfType, then call deserializeSpecified on the newly created vector
		static ValueVector *deserialize(InputStream *stream);
	};
	
	//! Helper class to cast to type T
	template<typename ST>
	struct TypeSerializer
	{
		//! write value v of type T to stream stored in type ST
		template<typename T>
		static void writeToStream(OutputStream *stream, T v)
		{
			ST sv = static_cast<ST>(v);
			stream->writeEndianIndependant(sv);
		}
		
		//! read value v of type from stream stored in type ST
		template<typename T>
		static void readFromStream(InputStream *stream, T *vRef)
		{
			ST sv = stream->readEndianIndependant<ST>();
			*vRef = static_cast<T>(sv);
		}
	};


	//! Helper  class to cast to type double, required because plateform specific representation of double can weired.
	template<>
	struct TypeSerializer<double>
	{
		//! write value v of type T to stream stored in type double
		template<typename T>
		static void writeToStream(OutputStream *stream, T v)
		{
			double sv = static_cast<double>(v);
			stream->writeDouble(sv);
		}
		
		//! read value v of type from stream stored in type double
		template<typename T>
		static void readFromStream(InputStream *stream, T *vRef)
		{
			double sv = stream->readDouble();
			*vRef = static_cast<T>(sv);
		}
	};
		
	//! A static struct templatized on a scalar variable T. It has a single member type, whose value is a ValueVector::Type constant corresponding to T. ST is the type used for serialisation
	template<typename T> struct TypeResolver { };
	//! A static struct templatized on a Int32. It has a single member type, whose value is the ValueVector::Type constant corresponding to Int32
	template<> struct TypeResolver<Int32> : public TypeSerializer<Int32> { static const ValueVector::Type type = ValueVector::INT; };
	//! A static struct templatized on a UInt32. It has a single member type, whose value is the ValueVector::Type constant corresponding to UInt32
	template<> struct TypeResolver<UInt32> : public TypeSerializer<UInt32> { static const ValueVector::Type type = ValueVector::UINT; };
	//! A static struct templatized on a Int16. It has a single member type, whose value is the ValueVector::Type constant corresponding to Int16
	template<> struct TypeResolver<Int16> : public TypeSerializer<Int16> { static const ValueVector::Type type = ValueVector::SHORT; };
	//! A static struct templatized on a UInt16. It has a single member type, whose value is the ValueVector::Type constant corresponding to UInt16
	template<> struct TypeResolver<UInt16> : public TypeSerializer<UInt16> { static const ValueVector::Type type = ValueVector::USHORT; };
	//! A static struct templatized on a Int8. It has a single member type, whose value is the ValueVector::Type constant corresponding to Int8
	template<> struct TypeResolver<Int8> : public TypeSerializer<Int8> { static const ValueVector::Type type = ValueVector::CHAR; };
	//! A static struct templatized on a UInt8. It has a single member type, whose value is the ValueVector::Type constant corresponding to Int8
	template<> struct TypeResolver<UInt8> : public TypeSerializer<UInt8> { static const ValueVector::Type type = ValueVector::UCHAR; };
	//! A static struct templatized on a float. It has a single member type, whose value is the ValueVector::Type constant corresponding to float
	template<> struct TypeResolver<float> : public TypeSerializer<float> { static const ValueVector::Type type = ValueVector::FLOAT; };
	//! A static struct templatized on a double. It has a single member type, whose value is the ValueVector::Type constant corresponding to double
	template<> struct TypeResolver<double> : public TypeSerializer<double> { static const ValueVector::Type type = ValueVector::DOUBLE; };
	//! A static struct templatized on a bool with as serialized value of Uint8. It has a single member type, whose value is the ValueVector::Type constant corresponding to bool
	template<> struct TypeResolver<bool> : public TypeSerializer<UInt8> { static const ValueVector::Type type = ValueVector::BOOL; };
	
	//! The template implementation of ValueVector
	template<typename T>
	struct ValueVectorTemplate: public ValueVector
	{
	public:
		//! The vector containing scalar values
		std::vector<T> value;
	
		//! Constructor. Create a ValueVectoTemplate<T> with size elements
		ValueVectorTemplate(Size size) { value.resize(size); }
	
		virtual void serialize(OutputStream *stream) const
		{
			stream->writeUInt32(TypeResolver<T>::type);
			serializeSpecified(stream);
		}
	
		virtual void serializeSpecified(OutputStream *stream) const
		{
			stream->writeUInt32(static_cast<UInt32>(value.size()));
			for (size_t i=0; i<value.size(); i++)
			{
				TypeResolver<T>::writeToStream(stream, value[i]);
			}
		}
	
		virtual void deserializeSpecified(InputStream *stream)
		{
			size_t size = stream->readUInt32();
			value.resize(size);
			for (size_t i=0; i<size; i++)
			{
				T v;
				TypeResolver<T>::readFromStream(stream, &v);
				value[i] = v;
			}
		}
	
		virtual std::string toString(unsigned index) const
		{
			std::ostringstream oss; oss << value.at(index);
			return oss.str();
		}
	
		virtual void fromString(unsigned index, const std::string &s)
		{
			std::istringstream iss(s);
			T v;
			iss >> v;
			value.at(index) = v;
		}
	
		virtual Size size(void) const
		{
			return value.size();
		}
		
		virtual Type getType(void) const
		{
			return TypeResolver<T>::type;
		}
	};
	
	//! For conveniance
	typedef ValueVectorTemplate<Int32> IntValueVector;
	//! For conveniance
	typedef ValueVectorTemplate<float> FloatValueVector;
	
	//! A ValueVectorTemplate<UInt8> that has nice methods to typecast to and from strings
	struct CharValueVector: public ValueVectorTemplate<UInt8>
	{
		//! Constructor. Create a CharValueVector with size characters
		CharValueVector(Size size = 0):ValueVectorTemplate<UInt8>(size) { }
		//! Constructor. Create a CharValueVector from string s
		CharValueVector(const std::string &s);
		//! Constructor. Create a CharValueVector from s
		CharValueVector(const char *s);
		//! Cast the CharValueVector to a string
		operator std::string() const;
	};
	
	
	// Base service handling
	// =====================
	
	//! A full description of a Service
	class ServiceDescription
	{
	public:
		//! The attributes of the service, each value is a bit and thus can be composed together to form a mask
		enum Flags
		{
			//! Service can only be read
			READ_ONLY=0x1,
			//! Each value has a name
			NAMED_VALUES=0x2,
			//! Each value is bounded
			CONSTRAINT_VALUES=0x4,
		};
	
		//! The name of the service
		CharValueVector name;
		//! The type of scalar values the service provides
		ValueVector::Type type;
		//! The attributes mask of the service
		UInt32 flags;
		//! The number of values in the service. If zero, service has dynamic size
		UInt32 length;
		//! If (flags & NAMED_VALUES) != 0, the vector containing the name for each value
		std::vector<CharValueVector> subnames;
		//! If (flags & CONSTRAINT_VALUES) != 0, the vector containing the minimum for each value
		ValueVector *mins;
		//! If (flags & CONSTRAINT_VALUES) != 0, the vector containing the maximum for each value
		ValueVector *maxs;
	
	public:
		//! Constructor
		ServiceDescription();
		//! Destructor
		~ServiceDescription();
		//! Serialize to a stream
		void serialize(OutputStream *stream);
		//! Deserialize from a stream
		void deserialize(InputStream *stream);
	};
	
	//! A service interface, can be read with getValues or written with setValues if (description.flags & READ_ONLY) == 0. User should subclass Service, fill description in the constructor and implement getValues and setValues
	class Service
	{
	public:
		//! The description of the service
		ServiceDescription description;
	
	public:
		//! Constructor
		Service() { }
		//! Destructor
		virtual ~Service() { }
	
		//! Read values from the service. Must be implemented by subclass
		virtual ValueVector *getValues(void) { return NULL; }
		//! Write values to the service. Must be implemented by subclass if (description.flags & READ_ONLY) == 0
		virtual void setValues(ValueVector *values) { }
	};
	
	//! Contains all the service a server provides
	class ServicesContainer
	{
	public:
		//! All services indexed by their UID
		std::map<UInt32, Service *> services;
		//! All UID indexed by the service they refer to
		std::map<Service *, UInt32> reverseServices;
		//! All services indexed by their name
		std::map<std::string, Service *> namedServices;
		//! All service UID are below maxUID
		UInt32 maxUID;
	
	public:
		//! Constructor
		ServicesContainer();
		//! Destructor, destroy all services
		virtual ~ServicesContainer();
		//! Add Service s
		void addService(Service *s);
		//! Remove service s and destroy it
		void eraseService(Service *s);
		//! Return the service given its name, return NULL if no service with this name is available
		Service *findService(const std::string &name);
	};
	
	//! ServicesServer is both a container for services and a TCP/IP server to allow network access to these services
	class ServicesServer: public NetworkServer, public ServicesContainer
	{
	protected:
		//! Called when datas are available on socket
		void incomingData(Socket *socket);
		//! Called when connection to client is closed on socket.
		void connectionClosed(Socket *socket);
		//! Called when a new connection has been established on socket.
		void incomingConnection(Socket *socket);
		
	public:
		//! Constructor, serve on port
		ServicesServer(UInt16 port);
		//! Destructor
		virtual ~ServicesServer();
	};
	
	
	// Network access handling
	// =======================
	
	//! An interface for answers from server
	class Answer
	{
	public:
		//! The type of answer
		enum Type
		{
			//! Error
			ERROR,
			//! List of services
			SERVICE_LIST,
			//! Values of a given service
			VALUES
		};
	
	public:
		//! Destructor
		virtual ~Answer() { }
		//! Serialize this answer to stream
		virtual void serialize(OutputStream *stream) = 0;
		//! Deserialize this answer from stream, the type has already been read because the Answer is already created
		virtual void deserializeSpecified(InputStream *stream) = 0;
		//! Deserialize an Answer from a stream and return it. First read the type, then create the corresponding Answer and call deserializeSpecified on it
		static Answer *deserialize(InputStream *stream);
	};
	
	//! An interface for calls to server
	class Call
	{
	public:
		//! The type of call
		enum Type
		{
			//! Hello call for connection
			HELLO=0,
			//! Service list request
			GET_SERVICE_LIST,
			//! Values read request
			GET_VALUES,
			//! Values write request
			SET_VALUES
		};
	
	public:
		//! Destructor
		virtual ~Call() { }
		//! Serialize this call to stream
		virtual void serialize(OutputStream *stream) = 0;
		//! Deserialize this call from stream, the type has already been read because the Call is already created
		virtual void deserializeSpecified(InputStream *stream) = 0;
		//! Deserialize a Call from a stream and return it. First read the type, then create the corresponding Call and call deserializeSpecified on it
		static Call *deserialize(InputStream *stream);
		//! Execute this call of the ServiceContainer registrar and return the corresponding answer. Return NULL if there is no answer
		virtual Answer *execute(ServicesContainer *registrar) { return NULL; }
	};
	
	//! Say hello at connection
	class HelloCall : public Call
	{
	public:
		//! Type of node connecting
		enum NodeType
		{
			//! Ishtar server node
			SERVER=0,
			//! Ishtar client node
			CLIENT
		} nodeType;
		//! The version of the protocol this node talks
		UInt32 version;
	
	public:
		//! Constructor
		HelloCall() { }
		//! Constructor
		HelloCall(NodeType t) { nodeType = t; version=PROTOCOL_VERSION; }
		//! Destructor
		virtual ~HelloCall() { }
		virtual void serialize(OutputStream *stream);
		virtual void deserializeSpecified(InputStream *stream);
	};
	
	//! Return an error to caller
	class ErrorAnswer: public Answer
	{
	public:
		//! The type of error
		enum ErrorTypes
		{
			//! A request service does not exists any more
			SERVICE_DOES_NOT_EXISTS = 0,
		} errorType;
	
	public:
		//! Constructor
		ErrorAnswer() { }
		//! Constructor
		ErrorAnswer(ErrorTypes error) { this->errorType = error; }
		//! Destructor
		virtual ~ErrorAnswer() { }
		virtual void serialize(OutputStream *stream);
		virtual void deserializeSpecified(InputStream *stream);
	};
	
	//! Request a list of all available services
	class GetServicesCall : public Call
	{
	public:
		//! Destructor
		virtual ~GetServicesCall() { }
		virtual void serialize(OutputStream *stream) { stream->writeUInt32(GET_SERVICE_LIST); }
		virtual void deserializeSpecified(InputStream *stream) { }
	
		virtual Answer *execute(ServicesContainer *registrar);
	};
	
	//! Return the description of all available services
	class ServiceAnswer: public Answer
	{
	public:
		//! A map of the UID and descriptions of services
		typedef std::map<UInt32, ServiceDescription *> DescriptionMap;
		//! The descriptions of all services
		DescriptionMap descriptions;
	
	public:
		//! Destructor
		virtual ~ServiceAnswer();
		//! To be called by client-side only, destroy the ServiceDescription in the map
		void deleteDescriptions(void);
		virtual void serialize(OutputStream *stream);
		virtual void deserializeSpecified(InputStream *stream);
	};
	
	//! Request the values of several services
	class GetValuesCall : public Call
	{
	public:
		//! The ID of the request, used by the proxy for queuing
		Int32 requestID;
		//! The UID of all services we want the values
		std::vector<UInt32> services;
	
	public:
		//! Constructor
		GetValuesCall() { requestID=0; }
		//! Destructor
		virtual ~GetValuesCall() { }
		virtual void serialize(OutputStream *stream);
		virtual void deserializeSpecified(InputStream *stream);
	
		virtual Answer *execute(ServicesContainer *registrar);
	};
	
	//! Return the values of several services
	class ValuesAnswer: public Answer
	{
	public:
		//! The ID of the request, used by the proxy for queuing
		Int32 requestID;
		//! A map of the UID and values of services
		typedef std::map<UInt32, ValueVector *> ValueVectorMap;
		//! The UID and values of several services
		ValueVectorMap values;
	
	public:
		//! Destructor
		virtual ~ValuesAnswer();
		//! To be called by client-side only, destroy the ValueVector in the map
		void deleteValues(void);
		virtual void serialize(OutputStream *stream);
		virtual void deserializeSpecified(InputStream *stream);
	};
	
	//! Set the values of several services
	class SetValuesCall : public Call
	{
	public:
		//! The UID and the values of all services we want to set
		std::map<UInt32, ValueVector *> values;
	
	public:
		//! Destructor
		virtual~SetValuesCall();
	
		//! Destroy the ValueVector in the map
		void deleteValues(void);
		virtual void serialize(OutputStream *stream);
		virtual void deserializeSpecified(InputStream *stream);
	
		virtual Answer *execute(ServicesContainer *registrar);
	};
}

#endif

